import numpy as np
import pypianoroll
import glob

def get_stats_dicts(results):
    s_stats = {
        "min": results[0][0],
        "avg": results[0][0], 
        "max": results[0][0]
        }
    c_stats = {
        "min": results[0][1],
        "avg": results[0][1], 
        "max": results[0][1]
        }
    p_stats = {
        "min": results[0][2],
        "avg": results[0][2], 
        "max": results[0][2]
        }
    u_stats = {
        "min": results[0][3],
        "avg": results[0][3], 
        "max": results[0][3]
        }
    d = list(filter(lambda x: x > 0., results[0][4]))
    d_stats = {
        "min": min(d),
        "avg": np.mean(d),
        "max": max(d)
    }
    span = max(results[0][5])[0] - min(results[0][5])[0]
    sp_stats = {
        "min": span,
        "avg": span,
        "max": span
    }
    idx = 0
    for i in range(1, len(results)):

        if s_stats["min"] > results[i][0]:
            s_stats["min"] = results[i][0]
        if s_stats["max"] < results[i][0]:
            s_stats["max"] = results[i][0]
        s_stats["avg"] += results[i][0]

        if c_stats["min"] > results[i][1]:
            c_stats["min"] = results[i][1]
        if c_stats["max"] < results[i][1]:
            c_stats["max"] = results[i][1]
        c_stats["avg"] += results[i][1]

        
        if p_stats["min"] > results[i][2]:
            p_stats["min"] = results[i][2]
        if p_stats["max"] < results[i][2]:
            p_stats["max"] = results[i][2]
        p_stats["avg"] += results[i][2]

        
        if u_stats["min"] > results[i][3]:
            u_stats["min"] = results[i][3]
        if u_stats["max"] < results[i][3]:
            u_stats["max"] = results[i][3]
        u_stats["avg"] += results[i][3]
        
        d = list(filter(lambda x: x > 0., results[i][4]))
        if d_stats["min"] > min(d):
            d_stats["min"] = min(d)
        if d_stats["max"] < max(d):
            d_stats["max"] = max(d)
        d_stats["avg"] += np.mean(d)

        span = max(results[i][5])[0] - min(results[i][5])[0]
        if sp_stats["min"] > span:
            sp_stats["min"] = span
        if sp_stats["max"] < span:
            sp_stats["max"] = span
        sp_stats["avg"] += span
    
    s_stats["avg"] = float(s_stats["avg"]/len(results))
    c_stats["avg"] = float(c_stats["avg"]/len(results))
    p_stats["avg"] = float(p_stats["avg"]/len(results))
    u_stats["avg"] = float(u_stats["avg"]/len(results))
    d_stats["avg"] = float(d_stats["avg"]/len(results))
    sp_stats["avg"] = float(sp_stats["avg"]/len(results))

    result_dict = {
        "singles": s_stats,
        "chords": c_stats,
        "pauses": p_stats,
        "uniques": u_stats,
        "durations": d_stats,
        "spans": sp_stats
    }

    return result_dict

def calculate_statistics(data):
    results = []
    counter = 0
    for phrase in data:
        single_notes = 0
        chords = 0
        pauses = 0
        prev = [-1]
        dur_arr = []
        durations = 0
        span = []
        for step in phrase:
            notes_played = np.argwhere(step * 1. > 0).flatten().tolist()
            if len(notes_played) == 0 and prev[0] is not -1:
                dur_arr.append(durations)
                prev = [-1]
                durations = 0
            
            elif len(notes_played) == 0 and prev[0] is -1:
                prev = [-1]
                durations = 0
            
            elif prev == notes_played:
                durations += 1
                prev = notes_played
            elif len(notes_played) > 0 and prev[0] is -1:
                durations += 1
                prev = notes_played
            elif prev != notes_played and len(notes_played) > 0 and prev[0] is not -1:
                dur_arr.append(durations)
                prev = notes_played
                durations = 1
        
            if len(notes_played) == 0:
                pauses += 1
                prev = [-1]
            
            elif len(notes_played) == 1:
                single_notes += 1
                span.append(notes_played)
            
            elif len(notes_played) > 1:
                chords += 1
                span.append(notes_played)
        if durations > 1:
            dur_arr.append(durations)
        unique_steps = len(np.unique(phrase, axis=0))
        counter += 1
        results.append([single_notes, chords, pauses, unique_steps, dur_arr, span])
    stat_results = get_stats_dicts(results)
    print("Stats calculated!")
    return stat_results, results


def load_data(filepath):

    data = np.load(filepath)

    drums = data[:, :, :, :,0]
    bass = data[:, :, :, :,1]
    guitar = data[:, :, :, :,2]
    strings = data[:, :, :, :,3]
    piano = data[:, :, :, :,4]

    drums = drums.reshape(len(drums), 384, 84)
    bass = bass.reshape(len(bass), 384, 84)
    guitar = guitar.reshape(len(guitar), 384, 84)
    strings = strings.reshape(len(strings), 384, 84)
    piano = piano.reshape(len(piano), 384, 84)

    data_dict = {
        "drums": drums,
        "bass": bass,
        "guitar": guitar,
        "strings": strings,
        "piano": piano
    }

    return data_dict

def load_midi_samples(filepath, num_samples):
    generated = []
    reals = []
    for i in range(num_samples):
        gen = pypianoroll.parse("{0}/gen_acc{1}_isolated.mid".format(filepath, i+1))
        gen = gen.tracks[0].pianoroll
        generated.append(gen)
        real = pypianoroll.parse("{0}/real_acc{1}_isolated.mid".format(filepath, i+1))
        real = real.tracks[0].pianoroll
        reals.append(real)
    return generated, reals


if __name__ == "__main__":
    data = []
    stats = []
    instr = []
    # Calculate statistics for generated samples
    load_samples = True
    if load_samples:
        path = "output"
        num_samples = len(glob.glob("{}/gen_acc*_isolated.mid".format(path)))
        generated, reals = load_midi_samples(path, num_samples)
        generated = np.array(generated) / 100.
        reals = np.array(reals) / 100.
        gen_stats, gen_results = calculate_statistics(generated)
        real_stats, real_results = calculate_statistics(reals)
        stats = ([gen_stats, real_stats])
        results = ([gen_results, real_results])
        instr = ["GENERATED", "REAL"]

    # Calculate statistics for training data
    else:
        filepath = 'data/training_data.npy'
        data = load_data(filepath)
        d_stats, d_results = calculate_statistics(data["drums"])
        b_stats, b_results = calculate_statistics(data["bass"])
        g_stats, g_results = calculate_statistics(data["guitar"])
        s_stats, s_results = calculate_statistics(data["strings"])
        p_stats, p_results = calculate_statistics(data["piano"])

        stats = ([d_stats, b_stats, g_stats, s_stats, p_stats])
        results = ([d_results, b_results, g_results, s_results, p_results])
        instr = ["DRUMS", "BASS", "GUITAR", "STRINGS", "PIANO"]
    
    
    for idx, stat in enumerate(stats):
        print("{} STATISTICS:".format(instr[idx]))
        print("-----------------------")
        print("Pauses: MAX = {0}, MIN = {1}, AVG = {2}".format(stat["pauses"]["max"], stat["pauses"]["min"], stat["pauses"]["avg"]))
        print("Chords: MAX = {0}, MIN = {1}, AVG = {2}".format(stat["chords"]["max"], stat["chords"]["min"], stat["chords"]["avg"]))
        print("Singles: MAX = {0}, MIN = {1}, AVG = {2}".format(stat["singles"]["max"], stat["singles"]["min"], stat["singles"]["avg"]))
        print("Uniques: MAX = {0}, MIN = {1}, AVG = {2}".format(stat["uniques"]["max"], stat["uniques"]["min"], stat["uniques"]["avg"]))
        print("Durations: MAX = {0}, MIN = {1}, AVG = {2}".format(stat["durations"]["max"], stat["durations"]["min"], stat["durations"]["avg"]))
        print("Spans: MAX = {0}, MIN = {1}, AVG = {2}".format(stat["spans"]["max"], stat["spans"]["min"], stat["spans"]["avg"]))
    
    # Save statistics
    #np.save("stats_raw", results)